/*
 * Driver for NAND support, Rick Bronson
 * borrowed heavily from:
 * (c) 1999 Machine Vision Holdings, Inc.
 * (c) 1999, 2000 David Woodhouse <dwmw2@infradead.org>
 *
 * Ported 'dynenv' to 'nand env.oob' command
 * (C) 2010 Nanometrics, Inc.
 * 'dynenv' -- Dynamic environment offset in NAND OOB
 * (C) Copyright 2006-2007 OpenMoko, Inc.
 * Added 16-bit nand support
 * (C) 2004 Texas Instruments
 *
 * Copyright 2010, 2012 Freescale Semiconductor
 * The portions of this file whose copyright is held by Freescale and which
 * are not considered a derived work of GPL v2-only code may be distributed
 * and/or modified under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 */
#include <common.h>
#include <linux/mtd/mtd.h>
#include <command.h>
#include <watchdog.h>
#include <malloc.h>
#include <asm/byteorder.h>
#include <jffs2/jffs2.h>
#include <nand.h>
#include <asm/arch/sys_flash_layout.h>
#include <asm/arch/system.h>

#if defined(CONFIG_CMD_MTDPARTS)

/* partition handling routines */
int mtdparts_init(void);
int id_parse(const char *id, const char **ret_id, u8 *dev_type, u8 *dev_num);
int find_dev_and_part(const char *id, struct mtd_device **dev,
		      u8 *part_num, struct part_info **part);
#endif
void nf_erase_block(struct mtd_info *rtk_mtd, int start_block, int block_num);
void nand_write_times_test(struct mtd_info *rtk_mtd,int start_block,int block_num);
static void nf_table_test(struct mtd_info *mtd,int block_addr,int test_way);
static void log2phy_test(struct mtd_info *mtd,int block_addr);
void nf_get_random_content(u_char *content, int content_size);
bool nf_compare_memory(u_char *src_mem,u_char *dst_mem, int memsize);
void dump_page_oob(struct mtd_info *rtk_mtd, int page);


#ifdef CONFIG_MTD_RBT_TEST
extern int test_erase_fail_remap;
extern int test_write_fail_remap;
extern int test_write_fail_bbt;
extern int test_read_warning_remap;
extern int test_read_fail_remap;
extern int check_buffer_empty(u_char *buff, int buf_size);
extern int load_table_from_nand(struct mtd_info *mtd,u_char tag, int is_rbt);
extern int update_bbt_region_to_nand(struct mtd_info *mtd,u_char * buf, int buf_len);
#endif
extern void dump_nf_ibt(struct mtd_info *mtd);
extern void dump_nf_bbt(struct mtd_info *mtd);
extern void dump_nf_rbt(struct mtd_info *mtd);
extern void dump_nf_bi_table(struct mtd_info *mtd);
extern int nf_get_phyblock(struct mtd_info *mtd,int logblock);
extern int nf_get_freeblock(struct mtd_info *mtd,int badblock);
extern int nf_read_page_content(struct mtd_info *mtd,unsigned int page_no,u_char * buf, unsigned int page_num, u_char tag,unsigned int skip_bad_page);
extern void mark_as_badblock(struct mtd_info *rtk_mtd,int block_id);
extern int rtk_block_isbad(struct mtd_info *mtd, u16 chipnr, loff_t ofs);
extern void nf_build_bootcode_log2phy_table(struct mtd_info *mtd);

void nf_test_bi_table(struct mtd_info *mtd);
void nf_test_erase_write_read(struct mtd_info *mtd,int block_addr);
void nf_test_remap(struct mtd_info *mtd,int block_addr,int way);
void nf_test_no_collect_table(struct mtd_info *mtd,int block_addr,int way);
void nf_dump_random(struct mtd_info *mtd,ulong start_addr, ulong length,int raw,char *content_buf);
int nf_write_random(struct mtd_info *mtd,char *buf,int start_addr, int length,int raw);
void nf_erase_flash(struct mtd_info *rtk_mtd);


static int nand_dump(nand_info_t *nand, ulong off, int only_oob, int repeat,int is_phy)
{
	int i;
	u_char *datbuf, *oobbuf, *p;
	static loff_t last;
	struct nand_chip *chip = nand->priv;

	if (repeat)
		off = last + nand->writesize;

	last = off;

	datbuf = malloc(nand->writesize);
	oobbuf = malloc(nand->oobsize);
	if (!datbuf || !oobbuf) {
		puts("No memory for page buffer\n");
		return 1;
	}
	off &= ~(nand->writesize - 1);
	loff_t addr = (loff_t) off;
	struct mtd_oob_ops ops;
	memset(&ops, 0, sizeof(ops));
	ops.datbuf = datbuf;
	ops.oobbuf = oobbuf;
	ops.len = nand->writesize;
	ops.ooblen = nand->oobsize;
	ops.mode = MTD_OOB_RAW;
	if(is_phy){
		unsigned int page = off >> chip->page_shift;
		i = chip->read_ecc_page(nand, 0, page, ops.datbuf, ops.oobbuf);
	}else{
		i = nand->read_oob(nand, addr, &ops);
	}
	
	if (i < 0) {
		printf("Error (%d) reading page %08lx\n", i, off);
		free(datbuf);
		free(oobbuf);
		return 1;
	}
	printf("Page %08lx dump:\n", off);
	i = nand->writesize >> 4;
	p = datbuf;

	while (i--) {
		if (!only_oob)
			printf("\t%02x %02x %02x %02x %02x %02x %02x %02x"
			       "  %02x %02x %02x %02x %02x %02x %02x %02x\n",
			       p[0], p[1], p[2], p[3], p[4], p[5], p[6], p[7],
			       p[8], p[9], p[10], p[11], p[12], p[13], p[14],
			       p[15]);
		p += 16;
	}
	puts("OOB:\n");
	i = nand->oobsize >> 3;
	p = oobbuf;
	while (i--) {
		printf("\t%02x %02x %02x %02x %02x %02x %02x %02x\n",
		       p[0], p[1], p[2], p[3], p[4], p[5], p[6], p[7]);
		p += 8;
	}
	free(datbuf);
	free(oobbuf);

	return 0;
}

/* ------------------------------------------------------------------------- */

static int set_dev(int dev)
{
	if (dev < 0 || dev >= CONFIG_SYS_MAX_NAND_DEVICE ||
	    !nand_info[dev].name) {
		puts("No such device\n");
		return -1;
	}

	if (nand_curr_device == dev)
		return 0;

	printf("Device %d: %s", dev, nand_info[dev].name);
	puts("... is now current device\n");
	nand_curr_device = dev;

#ifdef CONFIG_SYS_NAND_SELECT_DEVICE
	board_nand_select_device(nand_info[dev].priv, dev);
#endif

	return 0;
}

static inline int str2off(const char *p, loff_t *num)
{
	char *endptr;

	*num = simple_strtoull(p, &endptr, 16);
	return *p != '\0' && *endptr == '\0';
}

static inline int str2long(const char *p, ulong *num)
{
	char *endptr;

	*num = simple_strtoul(p, &endptr, 16);
	return *p != '\0' && *endptr == '\0';
}

static int get_part(const char *partname, int *idx, loff_t *off, loff_t *size)
{
#ifdef CONFIG_CMD_MTDPARTS
	struct mtd_device *dev;
	struct part_info *part;
	u8 pnum;
	int ret;

	ret = mtdparts_init();
	if (ret)
		return ret;

	ret = find_dev_and_part(partname, &dev, &pnum, &part);
	if (ret)
		return ret;

	if (dev->id->type != MTD_DEV_TYPE_NAND) {
		puts("not a NAND device\n");
		return -1;
	}

	*off = part->offset;
	*size = part->size;
	*idx = dev->id->num;

	ret = set_dev(*idx);
	if (ret)
		return ret;

	return 0;
#else
	puts("offset is not a number\n");
	return -1;
#endif
}

static int arg_off(const char *arg, int *idx, loff_t *off, loff_t *maxsize)
{
	if (!str2off(arg, off))
		return get_part(arg, idx, off, maxsize);

	if (*off >= nand_info[*idx].size) {
		puts("Offset exceeds device limit\n");
		return -1;
	}

	*maxsize = nand_info[*idx].size - *off;
	return 0;
}

static int arg_off_size(int argc, char *const argv[], int *idx,
			loff_t *off, loff_t *size)
{
	int ret;
	loff_t maxsize = 0;

	if (argc == 0) {
		*off = 0;
		*size = nand_info[*idx].size;
		goto print;
	}

	ret = arg_off(argv[0], idx, off, &maxsize);
	if (ret)
		return ret;

	if (argc == 1) {
		*size = maxsize;
		goto print;
	}

	if (!str2off(argv[1], size)) {
		printf("'%s' is not a number\n", argv[1]);
		return -1;
	}

	if (*size > maxsize) {
		puts("Size exceeds partition or device limit\n");
		return -1;
	}

print:
	printf("device %d ", *idx);
	if (*size == nand_info[*idx].size)
		puts("whole chip\n");
	else
		printf("offset 0x%lx, size 0x%lx\n",
		       (unsigned long)(*off), (unsigned long)(*size));
	return 0;
}

#ifdef CONFIG_CMD_NAND_LOCK_UNLOCK
static void print_status(ulong start, ulong end, ulong erasesize, int status)
{
	printf("%08lx - %08lx: %08lx blocks %s%s%s\n",
		start,
		end - 1,
		(end - start) / erasesize,
		((status & NAND_LOCK_STATUS_TIGHT) ?  "TIGHT " : ""),
		((status & NAND_LOCK_STATUS_LOCK) ?  "LOCK " : ""),
		((status & NAND_LOCK_STATUS_UNLOCK) ?  "UNLOCK " : ""));
}

static void do_nand_status(nand_info_t *nand)
{
	ulong block_start = 0;
	ulong off;
	int last_status = -1;

	struct nand_chip *nand_chip = nand->priv;
	/* check the WP bit */
	nand_chip->cmdfunc(nand, NAND_CMD_STATUS, -1, -1);
	printf("device is %swrite protected\n",
		(nand_chip->read_byte(nand) & 0x80 ?
		"NOT " : ""));

	for (off = 0; off < nand->size; off += nand->erasesize) {
		int s = nand_get_lock_status(nand, off);

		/* print message only if status has changed */
		if (s != last_status && off != 0) {
			print_status(block_start, off, nand->erasesize,
					last_status);
			block_start = off;
		}
		last_status = s;
	}
	/* Print the last block info */
	print_status(block_start, off, nand->erasesize, last_status);
}
#endif

#ifdef CONFIG_ENV_OFFSET_OOB
unsigned long nand_env_oob_offset;

int do_nand_env_oob(cmd_tbl_t *cmdtp, int argc, char *const argv[])
{
	int ret;
	uint32_t oob_buf[ENV_OFFSET_SIZE/sizeof(uint32_t)];
	nand_info_t *nand = &nand_info[0];
	char *cmd = argv[1];

	if (CONFIG_SYS_MAX_NAND_DEVICE == 0 || !nand->name) {
		puts("no devices available\n");
		return 1;
	}

	set_dev(0);

	if (!strcmp(cmd, "get")) {
		ret = get_nand_env_oob(nand, &nand_env_oob_offset);
		if (ret)
			return 1;

		printf("0x%08lx\n", nand_env_oob_offset);
	} else if (!strcmp(cmd, "set")) {
		loff_t addr;
		loff_t maxsize;
		struct mtd_oob_ops ops;
		int idx = 0;

		if (argc < 3)
			goto usage;

		if (arg_off(argv[2], &idx, &addr, &maxsize)) {
			puts("Offset or partition name expected\n");
			return 1;
		}

		if (idx != 0) {
			puts("Partition not on first NAND device\n");
			return 1;
		}

		if (nand->oobavail < ENV_OFFSET_SIZE) {
			printf("Insufficient available OOB bytes:\n"
			       "%d OOB bytes available but %d required for "
			       "env.oob support\n",
			       nand->oobavail, ENV_OFFSET_SIZE);
			return 1;
		}

		if ((addr & (nand->erasesize - 1)) != 0) {
			printf("Environment offset must be block-aligned\n");
			return 1;
		}

		ops.datbuf = NULL;
		ops.mode = MTD_OOB_AUTO;
		ops.ooboffs = 0;
		ops.ooblen = ENV_OFFSET_SIZE;
		ops.oobbuf = (void *) oob_buf;

		oob_buf[0] = ENV_OOB_MARKER;
		oob_buf[1] = addr / nand->erasesize;

		ret = nand->write_oob(nand, ENV_OFFSET_SIZE, &ops);
		if (ret) {
			printf("Error writing OOB block 0\n");
			return ret;
		}

		ret = get_nand_env_oob(nand, &nand_env_oob_offset);
		if (ret) {
			printf("Error reading env offset in OOB\n");
			return ret;
		}

		if (addr != nand_env_oob_offset) {
			printf("Verification of env offset in OOB failed: "
			       "0x%08llx expected but got 0x%08lx\n",
			       (unsigned long long)addr, nand_env_oob_offset);
			return 1;
		}
	} else {
		goto usage;
	}

	return ret;

usage:
	return CMD_RET_USAGE;
}

#endif

static void nand_info_print(nand_info_t *nand)
{
	struct nand_chip *this = (struct nand_chip *) nand->priv;

	printf("\nRTK NAND INFO:\n");
	printf("nand_PartNum:  %s\n", nand->PartNum);
	printf("nand_size:  %lu\n", (unsigned long)this->device_size);
	printf("chip_size:  %lu\n", (unsigned long)this->chipsize);
	printf("block_size:  %u\n", nand->erasesize);
	printf("page_size:  %u\n", nand->oobblock);
	printf("oob_size:  %u\n", nand->oobsize);
	printf("ppb:  %u\n", nand->erasesize/nand->oobblock);
	printf("RBA:  %u\n", this->RBA);
	printf("BBs:  %u\n\n", this->BBs);

	return;

}

static void nand_print_and_set_info(int idx)
{
	nand_info_t *nand = &nand_info[idx];
	struct nand_chip *chip = nand->priv;
	const int bufsz = 32;
	char buf[bufsz];

	printf("Device %d: ", idx);
	if (chip->numchips > 1)
		printf("%dx ", chip->numchips);
	printf("%s, sector size %u KiB\n",
	       nand->name, nand->erasesize >> 10);
	printf("  Page size  %8d b\n", nand->writesize);
	printf("  OOB size   %8d b\n", nand->oobsize);
	printf("  Erase size %8d b\n", nand->erasesize);

	/* Set geometry info */
	sprintf(buf, "%x", nand->writesize);
	setenv("nand_writesize", buf);

	sprintf(buf, "%x", nand->oobsize);
	setenv("nand_oobsize", buf);

	sprintf(buf, "%x", nand->erasesize);
	setenv("nand_erasesize", buf);

	nand_info_print(nand);
	return ;
}

static int raw_access(nand_info_t *nand, ulong addr, loff_t off, ulong count,
			int read)
{
	int ret = 0;

	while (count--) {
		/* Raw access */
		mtd_oob_ops_t ops = {
			.datbuf = (u8 *)addr,
			.oobbuf = ((u8 *)addr) + nand->writesize,
			.len = nand->writesize,
			.ooblen = nand->oobsize,
			.mode = MTD_OOB_RAW
		};

		if (read)
			ret = nand->read_oob(nand, off, &ops);
		else
			ret = nand->write_oob(nand, off, &ops);

		if (ret) {
			printf("%s: error at offset %llx, ret %d\n",
				__func__, (long long)off, ret);
			break;
		}

		addr += nand->writesize + nand->oobsize;
		off += nand->writesize;
	}

	return ret;
}

#ifdef CONFIG_CMD_NAND_PRESSURE_TEST
#define NOT_ERASE_ALIGNED(mtd, x) (x & (mtd->erasesize-1)) != 0
#define UBOOT_DDR_OFFSET  0xa0000000
#define TEMP_WRITE_ADDR		(0xa0200000-UBOOT_DDR_OFFSET)
#define TEMP_READ_ADDR		(0xa0400000-UBOOT_DDR_OFFSET)

static int nand_pressure_test(unsigned int s_addr, unsigned int e_addr, unsigned int times,unsigned char mode)
{
	int dev = nand_curr_device;
	nand_info_t *nand =&nand_info[dev];		//nand_info[0];
	//struct nand_chip *this = (struct nand_chip *) nand->priv;
	unsigned int i = 0, j=0;
	int rc = 0;
	unsigned int block_e, block_s;
	struct erase_info instr;
	size_t retlen;

	if (NOT_ERASE_ALIGNED (nand, s_addr) || NOT_ERASE_ALIGNED(nand, e_addr)) {
		printf ("Error: start_addr and end_addr must block_aligned!\n");
		return -1;
	}

	block_s = s_addr/nand->erasesize;
	block_e = e_addr/nand->erasesize;

	instr.mtd = nand;
	instr.len = nand->erasesize;

	if(mode){

		for(j=0; j<times; j++){
			for(i=block_s; i<block_e; i++ ){
				instr.addr = i*nand->erasesize;
				//erase
				rc = nand->erase(nand,&instr);
				if (rc != 0) {
					printf("\n[ERR] Erase failure at offset: 0x%x(block0x%x)\n",(unsigned int)instr.addr,i);
					return -1;
				}
				//write
				rc = nand->write_ecc(nand, instr.addr, instr.len, &retlen, (u_char *)TEMP_WRITE_ADDR,NULL,NULL);
				if((rc != 0)||(retlen != instr.len)){
					printf("\n[ERR] Write failure at offset: 0x%x(block0x%x)\n",(unsigned int)instr.addr,i);
					return -1;
				}

				memset(TEMP_READ_ADDR, 0x0, nand->erasesize);

				//read
				rc = nand->read(nand,  instr.addr,  instr.len, &retlen, (u_char *)TEMP_READ_ADDR);
				if((rc != 0)||(retlen != instr.len)){
					printf("\n[ERR] Read failure at offset: 0x%x(block0x%x)\n",(unsigned int)instr.addr,i);
					return -1;
				}
				//verify
				if(0 != memcmp((unsigned char *)TEMP_WRITE_ADDR, (unsigned char *)TEMP_READ_ADDR, nand->erasesize)){
					printf("\n[ERR] Verify failure at offset: 0x%x(block0x%x)\n",(unsigned int)instr.addr,i);
					return -1;
				}
			}

			{
				int percent_complete = -1;
				unsigned long long n = j * 100ULL;
				int percent;

				do_div(n, times);
				percent = (int)n;

				if (percent != percent_complete) {
					percent_complete = percent;

					printf("\rTest  %d times -- %3d%% complete.", j, percent);

				}
			}
		}

		printf("\rTest  %d times -- %3d%% complete.", times, 100);

	}else{
		for(i=block_s; i<block_e; i++){
			instr.addr = i*nand->erasesize;
			for(j=0; j<times; j++){
				//erase
				rc = nand->erase(nand,&instr);
				if (rc != 0) {
					printf("\n[ERR] Erase failure at offset: 0x%x(block0x%x)\n",(unsigned int)instr.addr,i);
					return -1;
				}
				//write
				rc = nand->write_ecc(nand, instr.addr, instr.len, &retlen, (u_char *)TEMP_WRITE_ADDR,NULL,NULL);
				if((rc != 0)||(retlen != instr.len)){
					printf("\n[ERR] Write failure at offset: 0x%x(block0x%x)\n",(unsigned int)instr.addr,i);
					return -1;
				}

				memset(TEMP_READ_ADDR, 0x0, nand->erasesize);

				//read
				rc = nand->read(nand,  instr.addr,  instr.len, &retlen, (u_char *)TEMP_READ_ADDR);
				if((rc != 0)||(retlen != instr.len)){
					printf("\n[ERR] Read failure at offset: 0x%x(block0x%x)\n",(unsigned int)instr.addr,i);
					return -1;
				}
				//verify
				if(0 != memcmp((unsigned char *)TEMP_WRITE_ADDR,(unsigned char *)TEMP_READ_ADDR, nand->erasesize)){
					printf("\n[ERR] Verify failure at offset: 0x%x(block0x%x)\n",(unsigned int)instr.addr,i);
					return -1;
				}
			}

			{
				int percent_complete = -1;
				unsigned long long n =(i-block_s) * 100ULL;
				int percent;

				do_div(n, (block_e -block_s));
				percent = (int)n;

				if (percent != percent_complete) {
					percent_complete = percent;

					printf("\rTest at 0x%lx -- %3d%% complete.",
				       		(unsigned long) instr.addr, percent);

				}
			}
		}

		printf("\rTest at 0x%lx -- %3d%% complete.",
				       		(unsigned long) instr.addr, 100);
	}
	printf("\n\nTest finish!\n");
	return 0;
}
#endif

void nf_erase_perf_test(struct mtd_info *mtd,unsigned int start_addr,unsigned int size)
{
	struct nand_chip *this=(struct nand_chip *)mtd->priv;
	unsigned long begin = 0,end = 0,spend_time = 0;
	unsigned int block_id = 0, block_start = 0,block_num = 0,rate = 0;
	block_start = start_addr / this->block_size;
	block_num   = size/this->block_size;
	
	begin = get_boot_time_us();    
    
	for(block_id = 0; block_id < block_num; block_id++)
	{
		if(this->erase_block(mtd,0,(block_start+block_id)*this->ppb) < 0){
			printf("%s %d erase block page %d failed! \n",__FUNCTION__,__LINE__,
				(block_start+block_id)*this->ppb);
		}
	}
	
	end  = get_boot_time_us();
	spend_time =  end-begin;
	rate = spend_time/block_num;
	
	printf("%s erase time begin %lu end %lu time %luus size 0x%x erase block time %5dus \n",
		__FUNCTION__,begin,end,spend_time,size,rate);
}

void nf_write_perf_test(struct mtd_info *mtd,unsigned int start_addr,unsigned int size)
{
	struct nand_chip *this = (struct nand_chip *)mtd->priv;
	unsigned long begin = 0,end = 0,spend_time = 0;
	unsigned int loop = 0,page_num = 0,rate = 0;
	unsigned char *page_content = (unsigned char *)malloc(1<<this->page_shift);
	size_t retlen = 0;
	loff_t to = 0;
	if(page_content == NULL){
		printf("malloc size 0x%x fail!\n",this->page_size);
		return;
	}
	nf_get_random_content(page_content,this->page_size);
	to = start_addr;
	page_num = size/this->page_size;
	
	begin = get_boot_time_us(); 
	
	for(loop = 0; loop < page_num; loop++)
	{
		if(mtd->write(mtd, to+this->page_size*loop, this->page_size, &retlen, 
			page_content) < 0){	
			break;
		}
	}
	
	end  = get_boot_time_us();
	spend_time = end-begin;
	rate = (size/1024)*1000000/spend_time;
	printf("%s write time begin %lu end %lu time %luus size 0x%x %5dKB/s\n",
		__FUNCTION__,begin,end,spend_time,size,rate);
	if(page_content != NULL){
		free(page_content);
	}
}

void nf_read_perf_test(struct mtd_info *mtd,unsigned int start_addr,unsigned int size)
{
	struct nand_chip *this = (struct nand_chip *)mtd->priv;
	unsigned long begin = 0,end = 0,spend_time = 0;
	unsigned int loop = 0,page_num = 0,page_start = 0,rate = 0,type = 0;
	unsigned char *page_content = NULL, *oob_content = NULL, *tmp_buf = NULL;
	size_t retlen = 0;
	
	if((page_content = (unsigned char *)malloc(1<<this->page_shift)) == NULL){
		printf("malloc size 0x%x fail!\n",this->page_size);
		return;
	}

	if((oob_content = (unsigned char *)malloc(this->oob_size)) == NULL){
		printf("malloc size 0x%x fail!\n",this->oob_size);
		goto end;
	}

	tmp_buf = (unsigned char *)TEMP_BUFFER_FOR_FLASH_DATA_ADDR;
	
	page_start = start_addr/this->page_size;
	page_num = size/this->page_size;

	for(type = 0; type < 2; type++)
	{
		begin = get_boot_time_us();  
		if(type == 0){
			mtd->read(mtd, start_addr, size, &retlen, tmp_buf);
		}else{
			for(loop = 0; loop < page_num;loop++)
				this->read_ecc_page(mtd,0,page_start+loop,page_content,oob_content);
		}		
		end  = get_boot_time_us();
		spend_time = end-begin;
		rate = (size/1024)*1000000/spend_time;
		printf("%s type %d read md begin %lu end %lu time %luus size 0x%x %6dKB/s\n",
			__FUNCTION__,type,begin,end,spend_time,size,rate);
	}

end:
	if(page_content != NULL){
		free(page_content);
	}
	
	if(oob_content != NULL){
		free(oob_content);
	}
}

void nf_performace_test(struct mtd_info *mtd)
{
	unsigned int start_addr = 0, size = 0;
	start_addr = SYS_PARAM_FACTORY_RO_START;
	size	   = SYS_PARAM_FACTORY_RO_SIZE;
	printf("%s start address 0x%x size 0x%x \n",__FUNCTION__,start_addr,size);
	/*erase performace test*/
	nf_erase_perf_test(mtd,start_addr,size);
	
	/*write performace test*/
	nf_write_perf_test(mtd,start_addr,size);

	/*read performace test*/
	nf_read_perf_test(mtd,start_addr,size);
}

int do_nand(cmd_tbl_t * cmdtp, int flag, int argc, char * const argv[])
{
	int i, ret = 0;
	ulong addr;
	loff_t off, size;
	char *cmd, *s;
	nand_info_t *nand;
#ifdef CONFIG_SYS_NAND_QUIET
	int quiet = CONFIG_SYS_NAND_QUIET;
#else
	int quiet = 0;
#endif
	const char *quiet_str = getenv("quiet");
	int dev = nand_curr_device = 0;
	int repeat = flag & CMD_FLAG_REPEAT;

	/* at least two arguments please */
	if (argc < 2)
		goto usage;

	if (quiet_str)
		quiet = simple_strtoul(quiet_str, NULL, 0) != 0;

	cmd = argv[1];

	/* Only "dump" is repeatable. */
	if (repeat && strcmp(cmd, "dump"))
		return 0;

	if (strcmp(cmd, "info") == 0) {

		putc('\n');
		for (i = 0; i < CONFIG_SYS_MAX_NAND_DEVICE; i++) {
			if (nand_info[i].name)
				nand_print_and_set_info(i);
		}
		return 0;
	}

	if (strcmp(cmd, "device") == 0) {
		if (argc < 3) {
			putc('\n');
			if (dev < 0 || dev >= CONFIG_SYS_MAX_NAND_DEVICE)
				puts("no devices available\n");
			else
				nand_print_and_set_info(dev);
			return 0;
		}

		dev = (int)simple_strtoul(argv[2], NULL, 10);
		set_dev(dev);

		return 0;
	}

#ifdef CONFIG_CMD_NAND_PRESSURE_TEST
	if (strcmp(cmd, "test") == 0) {

		unsigned int start_addr=0;
		unsigned int end_addr=0;
		unsigned int times=0;
		unsigned char mode = 0;

		start_addr = (unsigned int)simple_strtoul(argv[2], NULL, 16);
		end_addr = (unsigned int)simple_strtoul(argv[3], NULL, 16);
		times = (unsigned int)simple_strtoul(argv[4], NULL, 16);
		mode = (unsigned char)simple_strtoul(argv[5], NULL, 16);

		printf("start_addr = 0x%x  end_addr= 0x%x  times = 0x%x mode = %d\n\n",\
			start_addr, end_addr,times,mode);

		if((start_addr >= end_addr)||(times <= 0)||((mode !=0)&&(mode !=1))){
			printf("Argument Error! Please check it\n");
		}

		return  nand_pressure_test(start_addr, end_addr, times,mode);

	}
#endif

#ifdef CONFIG_ENV_OFFSET_OOB
	/* this command operates only on the first nand device */
	if (strcmp(cmd, "env.oob") == 0)
		return do_nand_env_oob(cmdtp, argc - 1, argv + 1);
#endif

	/* The following commands operate on the current device, unless
	 * overridden by a partition specifier.  Note that if somehow the
	 * current device is invalid, it will have to be changed to a valid
	 * one before these commands can run, even if a partition specifier
	 * for another device is to be used.
	 */
	if (dev < 0 || dev >= CONFIG_SYS_MAX_NAND_DEVICE ){ 
		puts("\nno devices available\n");
		return 1;
	}
	nand = &nand_info[dev];

	if (strcmp(cmd, "bad") == 0) {
		printf("\nDevice %d bad blocks:\n", dev);
		for (off = 0; off < nand->size; off += nand->erasesize)
			if (nand_block_isbad(nand, off)){
				unsigned int block_num;
				struct nand_chip *this = (struct nand_chip *) nand->priv;
				block_num = off >> this->phys_erase_shift;
				printf("  block_addr: 0x%08lx  ---> block_no: 0x%x\n", (unsigned long)off, block_num);
			}
		return 0;
	}

	if (strcmp(cmd, "status") == 0){
		unsigned char block_status = 0xff;
		unsigned int block_num;
		struct nand_chip *this = (struct nand_chip *) nand->priv;
		int i=0;

		printf("Bootcode area block status([xx, xx]-->[blocknum, blockstatus]):\n");
		for (off = 0; off < SYS_PARAM_BOOTCODE_SIZE; off += nand->erasesize){
			block_status = nand->block_status_get(nand, off);
			block_num = off >> this->phys_erase_shift;
			i++;

			printf("[0x%04x, 0x%02x]\t", block_num, block_status);
			if(i%8 == 0){
				printf("\n");
			}
		}
		return 0;

	}

	/*
	 * Syntax is:
	 *   0    1     2       3    4
	 *   nand erase [clean] [off size]
	 */
	if (strncmp(cmd, "erase", 5) == 0 || strncmp(cmd, "scrub", 5) == 0) {
		nand_erase_options_t opts;
		/* "clean" at index 2 means request to write cleanmarker */
		int clean = argc > 2 && !strcmp("clean", argv[2]);
		int scrub_yes = argc > 2 && !strcmp("-y", argv[2]);
		int o = (clean || scrub_yes) ? 3 : 2;
		int scrub = !strncmp(cmd, "scrub", 5);
		int spread = 0;
		int args = 2;
		const char *scrub_warn =
			"Warning: "
			"scrub option will erase all factory set bad blocks!\n"
			"         "
			"There is no reliable way to recover them.\n"
			"         "
			"Use this command only for testing purposes if you\n"
			"         "
			"are sure of what you are doing!\n"
			"\nReally scrub this NAND flash? <y/N>\n";

		if (cmd[5] != 0) {
			if (!strcmp(&cmd[5], ".spread")) {
				spread = 1;
			} else if (!strcmp(&cmd[5], ".part")) {
				args = 1;
			} else if (!strcmp(&cmd[5], ".chip")) {
				args = 0;
			} else {
				goto usage;
			}
		}

		/*
		 * Don't allow missing arguments to cause full chip/partition
		 * erases -- easy to do accidentally, e.g. with a misspelled
		 * variable name.
		 */
		if (argc != o + args)
			goto usage;

		printf("\nNAND %s: ", cmd);
		/* skip first two or three arguments, look for offset and size */
		if (arg_off_size(argc - o, argv + o, &dev, &off, &size) != 0)
			return 1;

		nand = &nand_info[dev];

		if(off == 0){
			printf("[Warning]Really erase block0x0?<y/N>\n");
			if (getc() == 'y'){
				printf("y\n");
			}else{
				return -1;
			}
		}

		//switch argument
		if(off < nand->size/nand->erasesize){
			printf("[Warning]Use block num as offset to excute erase operation?<y/N>\n");
			if (getc() == 'y'){
				printf("y\n");
				off = off * nand->erasesize;
			}
		}

		if(size< nand->oobblock){
			printf("[Warning]Use block num as size to excute erase operation?<y/N>\n");
			if (getc() == 'y'){
				printf("y\n");
				size = size * nand->erasesize;
			}
		}

		printf("[Info] Actual-offset 0x%lx, Actual-size 0x%lx\n",(unsigned long)off, (unsigned long)size);

		memset(&opts, 0, sizeof(opts));
		opts.offset = off;
		opts.length = size;
		opts.jffs2  = clean;
		opts.quiet  = quiet;
		opts.spread = spread;

		if (scrub) {
			if (!scrub_yes)
				puts(scrub_warn);

			if (scrub_yes)
				opts.scrub = 1;
			else if (getc() == 'y') {
				puts("y");
				if (getc() == '\r')
					opts.scrub = 1;
				else {
					puts("scrub aborted\n");
					return -1;
				}
			} else {
				puts("scrub aborted\n");
				return -1;
			}
		}
		ret = nand_erase_opts(nand, &opts);
		printf("%s\n", ret ? "ERROR" : "OK");

		return ret == 0 ? 0 : 1;
	}

	if (strncmp(cmd, "dump", 4) == 0) {
		char *cmd_tmp = NULL;
		int is_phy = 0;
		
		//sanity check
		cmd_tmp = argv[2];
		
		if (argc < 3)
			goto usage;
		
		if(strncmp(cmd_tmp, "phy", 3) == 0){
			off = (int)simple_strtoul(argv[3], NULL, 16);
			is_phy = 1;
		}else{
			off = (int)simple_strtoul(argv[2], NULL, 16);
			is_phy = 0;
		}

		ret = nand_dump(nand, off, !strcmp(&cmd[4], ".oob"), repeat, is_phy);
		
		return ret == 0 ? 1 : 0;
		
	}

	if (strncmp(cmd, "read", 4) == 0 || strncmp(cmd, "write", 5) == 0) {
		size_t rwsize;
		ulong pagecount = 1;
		int read;
		int raw;

		if (argc < 4)
			goto usage;

		addr = (ulong)simple_strtoul(argv[2], NULL, 16);

		read = strncmp(cmd, "read", 4) == 0; /* 1 = read, 0 = write */
		printf("\nNAND %s: ", read ? "read" : "write");

		nand = &nand_info[dev];

		s = strchr(cmd, '.');

		if (s && !strcmp(s, ".raw")) {
			raw = 1;

			if (arg_off(argv[3], &dev, &off, &size))
				return 1;

			if (argc > 4 && !str2long(argv[4], &pagecount)) {
				printf("'%s' is not a number\n", argv[4]);
				return 1;
			}

			if (pagecount * nand->writesize > size) {
				puts("Size exceeds partition or device limit\n");
				return -1;
			}

			rwsize = pagecount * (nand->writesize + nand->oobsize);
		} else {
			if (arg_off_size(argc - 3, argv + 3, &dev,
						&off, &size) != 0)
				return 1;

			rwsize = size;
		}
		printf("read/write offset = 0x%lx\n",(unsigned long)off);

		if (!s || !strcmp(s, ".jffs2") ||
		    !strcmp(s, ".e") || !strcmp(s, ".i")) {
			if (read)
				ret = nand_read_skip_bad(nand, off, &rwsize,
							 (u_char *)addr);
			else
				ret = nand_write_skip_bad(nand, off, &rwsize,
							  (u_char *)addr, 0);
#ifdef CONFIG_CMD_NAND_TRIMFFS
		} else if (!strcmp(s, ".trimffs")) {
			if (read) {
				printf("Unknown nand command suffix '%s'\n", s);
				return 1;
			}
			ret = nand_write_skip_bad(nand, off, &rwsize,
						(u_char *)addr,
						WITH_DROP_FFS);
#endif
#ifdef CONFIG_CMD_NAND_YAFFS
		} else if (!strcmp(s, ".yaffs")) {
			if (read) {
				printf("Unknown nand command suffix '%s'.\n", s);
				return 1;
			}
			ret = nand_write_skip_bad(nand, off, &rwsize,
						(u_char *)addr,
						WITH_INLINE_OOB);
#endif
		} else if (!strcmp(s, ".oob")) {
			/* out-of-band data */
			mtd_oob_ops_t ops = {
				.oobbuf = (u8 *)addr,
				.ooblen = rwsize,
				.mode = MTD_OOB_RAW
			};

			if (read)
				ret = nand->read_oob(nand, off, &ops);
			else
				ret = nand->write_oob(nand, off, &ops);
		} else if (raw) {
			ret = raw_access(nand, addr, off, pagecount, read);
		} else {
			printf("Unknown nand command suffix '%s'.\n", s);
			return 1;
		}

		printf(" %zu bytes %s: %s\n", rwsize,
		       read ? "read" : "written", ret ? "ERROR" : "OK");

		return ret == 0 ? 0 : 1;
	}

	if (strcmp(cmd, "markbad") == 0) {
		argc -= 2;
		argv += 2;

		if (argc <= 0)
			goto usage;

		while (argc > 0) {
			addr = simple_strtoul(*argv, NULL, 16);

			if (nand->block_markbad(nand, addr)) {
				printf("block 0x%08lx NOT marked "
					"as bad! ERROR %d\n",
					addr, ret);
				ret = 1;
			} else {
				printf("block 0x%08lx successfully "
					"marked as bad\n",
					addr);
			}
			--argc;
			++argv;
		}
		return ret;
	}

	if (strcmp(cmd, "biterr") == 0) {
		/* todo */
		return 1;
	}

#ifdef CONFIG_CMD_NAND_LOCK_UNLOCK
	if (strcmp(cmd, "lock") == 0) {
		int tight = 0;
		int status = 0;
		if (argc == 3) {
			if (!strcmp("tight", argv[2]))
				tight = 1;
			if (!strcmp("status", argv[2]))
				status = 1;
		}
		if (status) {
			do_nand_status(nand);
		} else {
			if (!nand_lock(nand, tight)) {
				puts("NAND flash successfully locked\n");
			} else {
				puts("Error locking NAND flash\n");
				return 1;
			}
		}
		return 0;
	}

	if (strcmp(cmd, "unlock") == 0) {
		if (arg_off_size(argc - 2, argv + 2, &dev, &off, &size) < 0)
			return 1;

		if (!nand_unlock(&nand_info[dev], off, size)) {
			puts("NAND flash successfully unlocked\n");
		} else {
			puts("Error unlocking NAND flash, "
			     "write and erase will probably fail\n");
			return 1;
		}
		return 0;
	}
#endif
	if (strcmp(cmd, "wp_test") == 0) {
		if (argc < 4) {
			putc('\n');
			puts("no enough command parameter\n");
			return 0;
		}
		int start_block = 0, block_num = 0;
		start_block = (int)simple_strtoul(argv[2], NULL, 10);
		block_num   = (int)simple_strtoul(argv[3], NULL, 10);
		nand_write_times_test(&nand_info[dev],start_block,block_num);
		return 0;
	}

	if (strcmp(cmd, "raw_erase") == 0) {
		if (argc < 4) {
			putc('\n');
			puts("no enough command parameter\n");
			return 0;
		}
		int start_block = 0, block_num = 0;
		start_block = (int)simple_strtoul(argv[2], NULL, 10);
		block_num   = (int)simple_strtoul(argv[3], NULL, 10);
		nf_erase_block(&nand_info[dev],start_block,block_num);
		return 0;
	}
	if (strcmp(cmd, "oob_read") == 0) {
		if (argc < 3) {
			putc('\n');
			puts("no enough command parameter\n");
			return 0;
		}
		int page = 0;
		page = (int)simple_strtoul(argv[2], NULL, 10);
	
		dump_page_oob(&nand_info[dev], page);
	
		return 0;
	}

	if (strcmp(cmd, "ibt_dump") == 0) {
		printf("enter dump_nf_bbt \n");
		dump_nf_ibt(&nand_info[dev]);
		return 0;
	}

	if (strcmp(cmd, "bbt_dump") == 0) {
		printf("enter dump_nf_bbt \n");
		dump_nf_bbt(&nand_info[dev]);
		return 0;
	}

	if (strcmp(cmd, "rbt_dump") == 0) {
		printf("enter dump_rbt \n");
		dump_nf_rbt(&nand_info[dev]);
		return 0;
	}

	if (strcmp(cmd, "bi_dump") == 0) {
		printf("enter dump_rbt \n");
		dump_nf_bi_table(&nand_info[dev]);
		return 0;
	}

	if (strcmp(cmd, "bit_test") == 0) {
		nf_test_bi_table(&nand_info[dev]);
		return 0;
	}

	if (strcmp(cmd, "pref_test") == 0){
		nf_performace_test(&nand_info[dev]);
		return 0;
	}

	if (strcmp(cmd, "table_test") == 0) {
		if (argc < 4) {
			putc('\n');
			puts("no enough command parameter\n");
			return 0;
		}
		int block_addr = 0, test_way = 0;
		block_addr = (int)simple_strtoul(argv[2], NULL, 10);
		test_way	= (int)simple_strtoul(argv[3], NULL, 10);
		nf_table_test(&nand_info[dev],block_addr,test_way);
	
		return 0;
	}

	if (strcmp(cmd, "log2phy_test") == 0) {
		if (argc < 3) {
			putc('\n');
			puts("no enough command parameter\n");
			return 0;
		}
		int block_addr = 0;
		block_addr = (int)simple_strtoul(argv[2], NULL, 10);
	
		log2phy_test(&nand_info[dev], block_addr);
	
		return 0;
	}

	if (strcmp(cmd, "ewr_test") == 0) {
		if (argc < 3) {
			putc('\n');
			puts("no enough command parameter\n");
			return 0;
		}
		int block_addr = 0;
		block_addr = (int)simple_strtoul(argv[2], NULL, 10);
	
		nf_test_erase_write_read(&nand_info[dev], block_addr);
	
		return 0;
	}
	
	if (strcmp(cmd, "test_remap") == 0) {
		if (argc < 4) {
			putc('\n');
			puts("no enough command parameter\n");
			return 0;
		}
		int block_addr = 0, test_way = 0;
		block_addr = (int)simple_strtoul(argv[2], NULL, 10);
		test_way = (int)simple_strtoul(argv[3], NULL, 10);
		nf_test_remap(&nand_info[dev],block_addr,test_way);
	
		return 0;
	}

	if (strcmp(cmd, "no_col_table") == 0) {
		if (argc < 4) {
			putc('\n');
			puts("no enough command parameter\n");
			return 0;
		}
		int block_addr = 0, test_way = 0;
		block_addr = (int)simple_strtoul(argv[2], NULL, 10);
		test_way	= (int)simple_strtoul(argv[3], NULL, 10);
		nf_test_no_collect_table(&nand_info[dev],block_addr,test_way);
		return 0;
	}

	if (strcmp(cmd, "random_dump") == 0) {
		if (argc < 5) {
			putc('\n');
			puts("no enough command parameter\n");
			return 0;
		}
		ulong addr = 0, size =0;
		int random = 0,dump_addr = 0;
		addr = (int)simple_strtoul(argv[2], NULL, 16);
		size	= (int)simple_strtoul(argv[3], NULL, 16);
		random	= (int)simple_strtoul(argv[4], NULL, 10);
		if(random == 3){
			dump_addr = (unsigned int)simple_strtoul(argv[5], NULL, 16);
		}
		nf_dump_random(&nand_info[dev],addr,size,random,(char *)dump_addr);
		return 0;
	}

	if (strcmp(cmd, "random_write") == 0) {
		if (argc < 5) {
			putc('\n');
			puts("no enough command parameter\n");
			return 0;
		}
		ulong addr = 0, size =0;
		int random = 0,dump_addr = 0;
		addr    = (int)simple_strtoul(argv[2], NULL, 16);
		size	= (int)simple_strtoul(argv[3], NULL, 16);
		random	= (int)simple_strtoul(argv[4], NULL, 10);
		dump_addr = (unsigned int)simple_strtoul(argv[5], NULL, 16);
		nf_write_random(&nand_info[dev],(char *)dump_addr,addr,size,random);
		return 0;
	}
	if (strcmp(cmd, "all_erase") == 0){
		nf_erase_flash(&nand_info[dev]);
		return 0;
	}
usage:
	return CMD_RET_USAGE;
}

U_BOOT_CMD(
	nand, CONFIG_SYS_MAXARGS, 1, do_nand,
	"NAND sub-system",
	"info - show available NAND devices\n"
	"nand device [dev] - show or set current device\n"
	"nand read - addr off|partition size\n"
	"nand write - addr off|partition size\n"
	"    read/write 'size' bytes starting at offset 'off'\n"
	"    to/from memory address 'addr', skipping bad blocks.\n"
	"nand read.raw - addr off|partition [count]\n"
	"nand write.raw - addr off|partition [count]\n"
	"    Use read.raw/write.raw to avoid ECC and access the flash as-is.\n"
#ifdef CONFIG_CMD_NAND_TRIMFFS
	"nand write.trimffs - addr off|partition size\n"
	"    write 'size' bytes starting at offset 'off' from memory address\n"
	"    'addr', skipping bad blocks and dropping any pages at the end\n"
	"    of eraseblocks that contain only 0xFF\n"
#endif
#ifdef CONFIG_CMD_NAND_YAFFS
	"nand write.yaffs - addr off|partition size\n"
	"    write 'size' bytes starting at offset 'off' with yaffs format\n"
	"    from memory address 'addr', skipping bad blocks.\n"
#endif
#ifdef CONFIG_CMD_NAND_PRESSURE_TEST
	"nand test start_addr end_addr times mode\n"
	"	test nand flash erase/write/read function\n"
	"	start_addr --> the start address of nand-flash that you want test\n"
	"	end_addr --> the end address of nand-flash that you want test\n"
	"	times -->erase/write/read test times\n"
	"	mode -->erase/write/read test mode(please choice 0 or 1)\n"
#endif
	"nand erase[.spread] [clean] off size - erase 'size' bytes "
	"from offset 'off'\n"
	"    With '.spread', erase enough for given file size, otherwise,\n"
	"    'size' includes skipped bad blocks.\n"
	"nand erase.part [clean] partition - erase entire mtd partition'\n"
	"nand erase.chip [clean] - erase entire chip'\n"
	"nand bad - show bad blocks\n"
	"nand status - show block status(bootcode-area)\n"
	"nand dump[.oob] [phy] off - dump page\n"
	"nand scrub [-y] off size | scrub.part partition | scrub.chip\n"
	"    really clean NAND erasing bad blocks (UNSAFE)\n"
	"nand markbad off [...] - mark bad block(s) at offset (UNSAFE)\n"
	"nand biterr off - make a bit error at offset (UNSAFE)"
#ifdef CONFIG_CMD_NAND_LOCK_UNLOCK
	"\n"
	"nand lock [tight] [status]\n"
	"    bring nand to lock state or display locked pages\n"
	"nand unlock [offset] [size] - unlock section"
#endif
#ifdef CONFIG_ENV_OFFSET_OOB
	"\n"
	"nand env.oob - environment offset in OOB of block 0 of"
	"    first device.\n"
	"nand env.oob set off|partition - set enviromnent offset\n"
	"nand env.oob get - get environment offset"
#endif
   	"\n"
   	"nand reinit                 - re-initialize nandflash driver\n"
   	"nand ip_test                - test nandflash ip\n"
   	"nand raw_read buf buf_size nand_offset  - read raw data\n"
   	"nand oob_read page  - read oob content\n"
   	"nand raw_write buf buf_size nand_offset  - write raw data\n"
   	"nand raw_erase start_block block_num  - erase block\n"
	"nand wp_test  start  number - start:start block id number: block number\n"
	"nand ibt_dump - dump ibt content \n"
	"nand bbt_dump - dump bbt content \n"
	"nand rbt_dump - dump rbt content \n"
	"nand bi_dump - dump bi table content \n"
	"nand random_dump addr size raw buf - raw(0:use map 1:no map 3:dump no map buf not print,need buf) \n"
	"nand random_write addr size raw buf - raw(0:use map 1:no map) \n"
	"nand bit_test - bitable test \n"
	"nand table_test  block_addr  test_way - block_addr:block id test_way:0:bbt 1:rbt\n"
	"nand log2phy_test block_addr - block_addr: block id \n"
	"nand ewr_test  block_addr - block_addr:erase write block id\n"
	"nand test_remap  block_addr  test_way - block_addr:block id test_way:0:read 1:write\n"
	"nand no_col_table  block_addr  test_way - block_addr:block id test_way:0:bbt 1:rbt\n"
	"nand pref_test                         - performace test \n"
	"nand all_erase					 erase whole flash \n"
);

static int nand_load_image(cmd_tbl_t *cmdtp, nand_info_t *nand,
			   ulong offset, ulong addr, char *cmd)
{
	int r;
	char *s;
	size_t cnt;
	image_header_t *hdr;
#if defined(CONFIG_FIT)
	const void *fit_hdr = NULL;
#endif

	s = strchr(cmd, '.');
	if (s != NULL &&
	    (strcmp(s, ".jffs2") && strcmp(s, ".e") && strcmp(s, ".i"))) {
		printf("Unknown nand load suffix '%s'\n", s);
		bootstage_error(BOOTSTAGE_ID_NAND_SUFFIX);
		return 1;
	}

	printf("\nLoading from %s, offset 0x%lx\n", nand->name, offset);

	cnt = nand->writesize;
	r = nand_read_skip_bad(nand, offset, &cnt, (u_char *) addr);
	if (r) {
		puts("** Read error\n");
		bootstage_error(BOOTSTAGE_ID_NAND_HDR_READ);
		return 1;
	}
	bootstage_mark(BOOTSTAGE_ID_NAND_HDR_READ);

	switch (genimg_get_format ((void *)addr)) {
	case IMAGE_FORMAT_LEGACY:
		hdr = (image_header_t *)addr;

		bootstage_mark(BOOTSTAGE_ID_NAND_TYPE);
		image_print_contents (hdr);

		cnt = image_get_image_size (hdr);
		break;
#if defined(CONFIG_FIT)
	case IMAGE_FORMAT_FIT:
		fit_hdr = (const void *)addr;
		puts ("Fit image detected...\n");

		cnt = fit_get_size (fit_hdr);
		break;
#endif
	default:
		bootstage_error(BOOTSTAGE_ID_NAND_TYPE);
		puts ("** Unknown image type\n");
		return 1;
	}
	bootstage_mark(BOOTSTAGE_ID_NAND_TYPE);

	r = nand_read_skip_bad(nand, offset, &cnt, (u_char *) addr);
	if (r) {
		puts("** Read error\n");
		bootstage_error(BOOTSTAGE_ID_NAND_READ);
		return 1;
	}
	bootstage_mark(BOOTSTAGE_ID_NAND_READ);

#if defined(CONFIG_FIT)
	/* This cannot be done earlier, we need complete FIT image in RAM first */
	if (genimg_get_format ((void *)addr) == IMAGE_FORMAT_FIT) {
		if (!fit_check_format (fit_hdr)) {
			bootstage_error(BOOTSTAGE_ID_NAND_FIT_READ);
			puts ("** Bad FIT image format\n");
			return 1;
		}
		bootstage_mark(BOOTSTAGE_ID_NAND_FIT_READ_OK);
		fit_print_contents (fit_hdr);
	}
#endif

	/* Loading ok, update default load address */

	load_addr = addr;

	return bootm_maybe_autostart(cmdtp, cmd);
}

int do_nandboot(cmd_tbl_t * cmdtp, int flag, int argc, char * const argv[])
{
	char *boot_device = NULL;
	int idx;
	ulong addr, offset = 0;
#if defined(CONFIG_CMD_MTDPARTS)
	struct mtd_device *dev;
	struct part_info *part;
	u8 pnum;

	if (argc >= 2) {
		char *p = (argc == 2) ? argv[1] : argv[2];
		if (!(str2long(p, &addr)) && (mtdparts_init() == 0) &&
		    (find_dev_and_part(p, &dev, &pnum, &part) == 0)) {
			if (dev->id->type != MTD_DEV_TYPE_NAND) {
				puts("Not a NAND device\n");
				return 1;
			}
			if (argc > 3)
				goto usage;
			if (argc == 3)
				addr = simple_strtoul(argv[1], NULL, 16);
			else
				addr = CONFIG_SYS_LOAD_ADDR;
			return nand_load_image(cmdtp, &nand_info[dev->id->num],
					       part->offset, addr, argv[0]);
		}
	}
#endif

	bootstage_mark(BOOTSTAGE_ID_NAND_PART);
	switch (argc) {
	case 1:
		addr = CONFIG_SYS_LOAD_ADDR;
		boot_device = getenv("bootdevice");
		break;
	case 2:
		addr = simple_strtoul(argv[1], NULL, 16);
		boot_device = getenv("bootdevice");
		break;
	case 3:
		addr = simple_strtoul(argv[1], NULL, 16);
		boot_device = argv[2];
		break;
	case 4:
		addr = simple_strtoul(argv[1], NULL, 16);
		boot_device = argv[2];
		offset = simple_strtoul(argv[3], NULL, 16);
		break;
	default:
#if defined(CONFIG_CMD_MTDPARTS)
usage:
#endif
		bootstage_error(BOOTSTAGE_ID_NAND_SUFFIX);
		return CMD_RET_USAGE;
	}
	bootstage_mark(BOOTSTAGE_ID_NAND_SUFFIX);

	if (!boot_device) {
		puts("\n** No boot device **\n");
		bootstage_error(BOOTSTAGE_ID_NAND_BOOT_DEVICE);
		return 1;
	}
	bootstage_mark(BOOTSTAGE_ID_NAND_BOOT_DEVICE);

	idx = simple_strtoul(boot_device, NULL, 16);

	if (idx < 0 || idx >= CONFIG_SYS_MAX_NAND_DEVICE || !nand_info[idx].name) {
		printf("\n** Device %d not available\n", idx);
		bootstage_error(BOOTSTAGE_ID_NAND_AVAILABLE);
		return 1;
	}
	bootstage_mark(BOOTSTAGE_ID_NAND_AVAILABLE);

	return nand_load_image(cmdtp, &nand_info[idx], offset, addr, argv[0]);
}

U_BOOT_CMD(nboot, 4, 1, do_nandboot,
	"boot from NAND device",
	"[partition] | [[[loadAddr] dev] offset]"
);

void nf_dump_byte(char *buf, int buf_size)
{
	int loop = 0;
	printf("buf 0x%x size 0x%x", (unsigned int)buf, buf_size);
	for(loop = 0; loop < buf_size; loop++)
	{
		if(loop%16 == 0)
		{
			printf("\n%08xh:",loop); 
		}
		printf(" %02x",*buf);
		buf++;
	}
	printf("\n");	   
}

void dump_page_oob(struct mtd_info *rtk_mtd, int page)
{
	struct nand_chip *this = (struct nand_chip *)rtk_mtd->priv;
	unsigned char *oob_buf = NULL;
	printf("%s %d nand block %d page %d !\n",
		__FUNCTION__,__LINE__,page/this->ppb,page);
	if((oob_buf = (unsigned char *)malloc(sizeof(this->oob_size))) == NULL){
		printf("%s %d malloc oob buf fail 0x%x!\n",
			__FUNCTION__,__LINE__,this->oob_size);
		return;
	}
	memset(oob_buf,0xFF,this->oob_size);
	this->read_oob(rtk_mtd,0,page,this->oob_size,oob_buf);
	nf_dump_byte((char *)oob_buf,this->oob_size);
	if(oob_buf != NULL){
		free(oob_buf);
	}
}

void nf_erase_block(struct mtd_info *rtk_mtd, int start_block, int block_num)
{
	struct nand_chip *this = (struct nand_chip *)rtk_mtd->priv;
	int loop = 0;
	printf("%s %d erase start %d number %d!\n",
		__FUNCTION__,__LINE__,start_block,block_num); 
	for(loop = 0; loop < block_num; loop++)
	{
		if (this->erase_block (rtk_mtd, 0, (start_block+loop)*this->ppb) == 0){
			printf("%s %d erase blockID %d sucessful!\n",
				__FUNCTION__,__LINE__,(start_block+loop)); 
		}else{
			printf("%s %d erase blockID %d failed!\n",
				__FUNCTION__,__LINE__,(start_block+loop)); 
		}
	}
}

bool nf_writeblock(struct mtd_info *rtk_mtd, int block_id, const u_char *block_content)
{
    	bool ret = true;
	struct nand_chip *this = (struct nand_chip *)rtk_mtd->priv;
	int  start_page        = block_id*this->ppb, page_id = 0; 
	const u_char *write_content    = block_content;
	/*get spare status*/
	if (this->isLastPage){
		page_id = block_id*this->ppb + this->ppb-1;
	}
	if ( this->read_oob(rtk_mtd,0, page_id, this->oob_size, this->g_oobbuf)!= 0 ){
		printf("%s %d Read error blockID %d page_id %d is bad block!\n",
			__FUNCTION__,__LINE__,block_id,page_id); 
	}
	
    	/*erase block */
	if (this->erase_block (rtk_mtd, 0, start_page) == 0){
		/* write block content*/
		for(page_id = start_page; page_id < (start_page+this->ppb); page_id++)
		{
		    if(this->write_ecc_page(rtk_mtd,0,page_id,write_content,this->g_oobbuf)== 0){
			    write_content += (1<<this->page_shift);
			}else{
			    printk("[%s %d ]write block %d page %d failure!!\n", __FUNCTION__,
			    		__LINE__,block_id,page_id);
				ret = false;
			    break;
			}
		}
		if(ret == true){
		    printk("[%s %d]write block %d successful!!\n", __FUNCTION__,__LINE__,block_id);
		}
	}else{   
		printk("[%s %d]erase block %d failure!!\n", __FUNCTION__,__LINE__,block_id);
		ret = false;
	}

    	/*mark as bad block*/
	if(ret == false){
	   mark_as_badblock(rtk_mtd,block_id);	
	}
	return ret;
}

bool nf_readblock(struct mtd_info *rtk_mtd, int block_id, u_char *block_content)
{
    bool ret = true;
	struct nand_chip *this = (struct nand_chip *)rtk_mtd->priv;
	int  start_page = block_id*this->ppb, page_id = 0;
	for(page_id = start_page; page_id < (start_page+this->ppb); page_id++)
	{
	    if(this->read_ecc_page(rtk_mtd, 0, page_id, block_content, this->g_oobbuf) == 0){
		    block_content += (1<<this->page_shift);
		}else{
		    printk("[%s]block %d page %d failure!!\n", __FUNCTION__,block_id,page_id);
			ret = false;
			break;
		}
	}

	/*mark as bad block*/
	if(ret == false){
		mark_as_badblock(rtk_mtd,block_id);
	}else{
		printk("[%s %d] block %d successful!!\n", __FUNCTION__,__LINE__,block_id);
	}

	return ret;
    
}

bool nf_compare_memory(u_char *src_mem,u_char *dst_mem, int memsize)
{
    bool ret = true;
	int loop = 0;
	for(loop = 0; loop< memsize; loop++)
	{
	    if((*src_mem) != (*dst_mem)){
		    printk("[%s]loop%d Src 0x%x dest 0x%x !!\n", __FUNCTION__,loop,*src_mem,*dst_mem); 
		    ret = false;
		}
		src_mem++;
		dst_mem++;
	}
	return ret;
}

void nf_get_random_content(u_char *content, int content_size)
{
    int loop = 0;
    for(loop = 0; loop < content_size; loop=loop+2)
	{
	    *content++ = 0x52;
		*content++ = 0x33;
	}
}

void single_block_test(struct mtd_info *rtk_mtd, int block_id)
{
    	struct nand_chip *this = (struct nand_chip *)rtk_mtd->priv;
	int  write_times = 0;
    	/*Choose good random block*/
	u_char *block_content     =  (u_char *)malloc(1<<this->phys_erase_shift);
	u_char *bak_block_content =  (u_char *)malloc(1<<this->phys_erase_shift);
	printk("[%s]block %d size 0x%x!!\n", __FUNCTION__,block_id,(1<<this->phys_erase_shift));
	if((block_content != NULL)&&(bak_block_content != NULL)){
	    nf_get_random_content(block_content,1<<this->phys_erase_shift);
	    while(true)
		{
		    write_times++;
			/*Write patten to nandflash block*/
		    if(nf_writeblock(rtk_mtd, block_id, block_content) == true){
			    /*read patten to compare*/
			    if(nf_readblock(rtk_mtd, block_id,bak_block_content) == false){
			        printk("[%s %d]block %d write_times %d readblock error !!\n", 
					__FUNCTION__,__LINE__,block_id,write_times);
			        break;
			    }else{
					if(nf_compare_memory(block_content,bak_block_content,1<<this->phys_erase_shift) == true){
						printk("[%s %d]block %d write_times %d compare OK !!\n", 
							__FUNCTION__,__LINE__,block_id,write_times);
					}else{
						printk("[%s %d]block %d write_times %d compare error !!\n", 
							__FUNCTION__,__LINE__,block_id,write_times);
						break;
					}
			    }
			}else{
			    printk("[%s %d]block %d write_times %d writeblock error !!\n", 
			    	__FUNCTION__,__LINE__,block_id,write_times);
			    break;
			}
		}
		free(block_content);
		free(bak_block_content);
	}
	
}

void nf_get_log2phy(struct mtd_info *rtk_mtd,int *log_phy, int start_block, int block_num)
{
	struct nand_chip *this = (struct nand_chip *)rtk_mtd->priv;
	int loop = 0,block_id = start_block;
	while(loop < block_num)
	{
		if(rtk_block_isbad(rtk_mtd,0,block_id*this->block_size) == false){
			log_phy[loop] = block_id;
			printk("[%s]loop%d phy %d!!\n", __FUNCTION__,loop,log_phy[loop]);
			loop++;
		}
		block_id++;	
	}
}

int multi_block_test(struct mtd_info *rtk_mtd, int start_block, int block_num,
	unsigned int test_times)
{
	struct nand_chip *this = (struct nand_chip *)rtk_mtd->priv;
	int block_id = 0,loop = 0, write_times = 0,start_id = 0,res = 1;
	bool ret = true;
	int content_size = (1<<this->phys_erase_shift);
	int  *log_phy           = (int *)malloc(sizeof(int) * block_num);	
	u_char *block_content     = (u_char *)malloc(content_size);
	u_char *bak_block_content = (u_char *)malloc(content_size);
	if((log_phy != NULL)&&(block_content != NULL)&&(bak_block_content != NULL)){
		/*get logical to physical table*/
		start_id = start_block;
		nf_get_log2phy(rtk_mtd,log_phy,start_id,block_num);
        
		while(true)
		{
			write_times++;
			/*write multi block*/
			while(loop < block_num )
			{
				block_id = log_phy[loop];
				memset(block_content,0,content_size);
				nf_get_random_content(block_content,content_size);
				if(nf_writeblock(rtk_mtd, block_id, block_content) == false){
					ret = false;
					printk("[%s %d]block number%d write_times %d writeblock error !!\n", 
						__FUNCTION__,__LINE__,block_num,write_times);
					break;
				}
				loop++;
			}
			if(ret == false){
				res = 0;
				break;
			}

			/*read multi block to compare*/
			loop = 0;
			while(loop < block_num)
			{
				block_id = log_phy[loop];
				memset(bak_block_content,0,content_size);
				if(nf_readblock(rtk_mtd, block_id, bak_block_content) == false){
					ret = false;
					printk("[%s %d]block number%d write_times %d readblock error !!\n", 
						__FUNCTION__,__LINE__,block_num,write_times);
					break;
				}else{
					if(nf_compare_memory(block_content,bak_block_content,content_size) == false){
						ret = false;
						printk("[%s %d]block id %d number%d write_times %d compare error !!\n", 
							__FUNCTION__,__LINE__,block_id+loop,block_num,write_times);
						break;
					}
				}
				loop++;
			}
			
			if(ret == false){
				res = 0;
				break;
			}else{
				printk("[%s %d]start_block %d block number%d write_times %d compare Successful !!\n", 
					__FUNCTION__,__LINE__,start_block,block_num,write_times);
			}
			loop = 0;

			if((test_times != 0xFFFFFFFF)&&(test_times == write_times)){
				break;
			}
			
		}
	}else{
	    printk("[%s %d]malloc buffer error size 0x%x !!\n", __FUNCTION__,__LINE__,content_size);
	}

	if(log_phy!= NULL){
		free(log_phy);
	}

	if(block_content != NULL){
		free(block_content);	
	}

	if(bak_block_content != NULL){
		free(bak_block_content);
	}
	return res;
}

void nand_write_times_test(struct mtd_info *rtk_mtd,int start_block,int block_num)
{	
	/*Continuity block Write Test*/
	multi_block_test(rtk_mtd,start_block,block_num,0xFFFFFFFF);

	/*Random block Write Test*/
	single_block_test(rtk_mtd,start_block+block_num+1);
}

int qc_nand_write_test(int start_block,int block_num,unsigned int test_times)
{
	nand_info_t *nand = &nand_info[0];
	return multi_block_test(nand,start_block,block_num,test_times);
}

static void log2phy_test(struct mtd_info *mtd,int block_addr)
{
	int phy_block = nf_get_phyblock(mtd,block_addr);
	printf("%s %d log %d phy %d \n",__FUNCTION__,__LINE__,block_addr,phy_block);
}

int nf_get_log2phy_page(struct mtd_info *mtd,int log_page)
{
	int phy_page = 0,log_block = 0,page_offset = 0;
	struct nand_chip *this = (struct nand_chip *)mtd->priv;
	log_block   = log_page/this->ppb;
	page_offset = log_page%this->ppb;
	phy_page    = nf_get_phyblock(mtd,log_block)*this->ppb+page_offset;
	printf("[log_page:%d,phy_page:%d]\n",log_page,phy_page);
	return phy_page;
}

int nf_read_random(struct mtd_info *mtd,char *buf,
	int start_addr, int length,int raw)
{
    int ret = 0,page_num = 0,loop = 0,size = length;
    struct nand_chip *this = (struct nand_chip *)mtd->priv;
    int left = start_addr%this->page_size,start_page = 0,phy_page = 0;
    char * tmp_buf = (char *)kmalloc(sizeof(char)*this->page_size,GFP_KERNEL);
    printf("%s %d start_addr 0x%x length 0x%x left 0x%x raw %d\n",__FUNCTION__,
    	__LINE__,start_addr,length,left,raw);
    if(tmp_buf!=NULL){
        if(left != 0){
            start_page = start_addr/this->page_size;
            size  = length>(this->page_size - left)?(this->page_size - left):length;
            phy_page = (raw == 0)?(nf_get_log2phy_page(mtd,start_page)):start_page;   
            printf("%s %d start_page %d left-size 0x%x  log_page %d phy_page %d\n",
               __FUNCTION__,__LINE__,start_page,size,start_page,phy_page); 
            if((ret = nf_read_page_content(mtd,phy_page,(unsigned char *)tmp_buf,
	    	1,0xFF,1))!=0){
                goto out;
            }
            memcpy(buf,tmp_buf+left,size);
            start_addr += size;
            buf    += size;
            length -=size;
        }
    
        start_page = start_addr/this->page_size;
        page_num   = length/this->page_size;
        printf("%s %d start_addr 0x%x start_page %d length 0x%x page_num %d\n",
		__FUNCTION__,__LINE__,start_addr,start_page,length,page_num); 
        for(loop = 0; loop < page_num ; loop++)
        {
            phy_page = (raw == 0)?(nf_get_log2phy_page(mtd,start_page)):start_page; 
            if((ret = nf_read_page_content(mtd,phy_page,(unsigned char *)tmp_buf,
	    	1,0xFF,1))!=0){
                goto out;
            }
            memcpy(buf,tmp_buf,this->page_size);
            buf += this->page_size;
            length -= this->page_size;
            start_page++;
        }
        if(length != 0){
            phy_page = (raw == 0)?(nf_get_log2phy_page(mtd,start_page)):start_page; 
            printf("%s %d start_page %d remain-size 0x%x  phy_page %d length 0x%x\n",
               __FUNCTION__,__LINE__,start_page,size,phy_page,length); 
            if((ret = nf_read_page_content(mtd,phy_page,(unsigned char *)tmp_buf,
	    	1,0xFF,1))!=0){
                goto out;
            }
            memcpy(buf,tmp_buf,length); 
        }
        kfree(tmp_buf);
    }
out:
    return ret;
}

void nf_dump_random(struct mtd_info *mtd,ulong start_addr, ulong length,int raw,
	char *content_buf)
{
    struct nand_chip *this = (struct nand_chip *)mtd->priv;
    char *buf = (char *)malloc(sizeof(char)*this->page_size);
    int dump_len = 0;
    printf("%s %d dump  start_addr %lu size %lu raw %d \n",
    	__FUNCTION__,__LINE__,start_addr,length,raw);	
    if(buf != NULL)
    {
    	  while(length > 0)
    	  {
    	      dump_len = length > this->page_size ? this->page_size:length;
              if(nf_read_random(mtd, buf,start_addr,dump_len,raw) == 0){
                  if((raw == 3)&&(content_buf != NULL)){
                      memcpy(content_buf,buf,dump_len);
                      content_buf = content_buf + dump_len;
                  }else{
                      nf_dump_byte(buf,dump_len);
                  }
                  
                  start_addr  += dump_len;
                  length      -= dump_len;    
              }else{
        	      printf("%s %d dump read error start_addr %lu size %lu \n",
        	       __FUNCTION__,__LINE__,start_addr,length);
                  break;
              }
        }
        free(buf);
    }else{
        printf("%s %d can't malloc 0x%x size \n",__FUNCTION__,__LINE__,this->page_size);
    }
}

int nf_write_random(struct mtd_info *mtd,char *buf,
	int start_addr, int length,int raw)
{
	int ret = 0,block_num = 0,loop = 0,size = length;
	struct nand_chip *this = (struct nand_chip *)mtd->priv;
	int left = start_addr%this->block_size,start_page = 0,phy_page = 0;
	char * tmp_buf = (char *)kmalloc(sizeof(char)*this->block_size,GFP_KERNEL);
	if(tmp_buf == NULL){
		printf("%s %d malloc buffer fail \n",__FUNCTION__,__LINE__);
		return -1;
	}
	printf("%s %d start_addr 0x%x length 0x%x left 0x%x raw %d\n",
		__FUNCTION__,__LINE__,start_addr,length,left,raw);
	if(left != 0){
		start_page = start_addr/this->page_size;
		size  = length>(this->block_size - left)?(this->block_size - left):length;
		phy_page = (raw == 0)?(nf_get_log2phy_page(mtd,start_page)):start_page;   
		printf("%s %d start_page %d left-size 0x%x  log_page %d phy_page %d\n",
			__FUNCTION__,__LINE__,start_page,size,start_page,phy_page);
		if(nf_readblock(mtd,phy_page/this->ppb,
			(unsigned char *)tmp_buf) == false){
			goto out;
		}
		memcpy(tmp_buf+left,buf,size);
		if(nf_writeblock(mtd,phy_page/this->ppb,
			(unsigned char *)tmp_buf) == false) {
			goto out;
		}
		start_addr += size;
		buf    += size;
		length -=size;
        }
	start_page  = start_addr/this->page_size;
	block_num   = length/this->block_size;   
	printf("%s start_page %d start_addr 0x%x,length 0x%x block_num %d \n",
			__FUNCTION__,start_page,start_addr,length,block_num); 
        for(loop = 0; loop < block_num ; loop++)
        {
		phy_page = (raw == 0)?(nf_get_log2phy_page(mtd,start_page)):start_page; 
		if(nf_writeblock(mtd,phy_page/this->ppb,(unsigned char *)buf) == false) {
			goto out;
		}
		buf += this->block_size;
		length -= this->block_size;
		start_page+= this->ppb;
        }
	
        if(length != 0){
		phy_page = (raw == 0)?(nf_get_log2phy_page(mtd,start_page)):start_page; 
		printf("%s start_page %d remain-size 0x%x  phy_page %d length 0x%x\n",
			__FUNCTION__,start_page,size,phy_page,length); 
		if(nf_readblock(mtd,phy_page/this->ppb,(unsigned char *)tmp_buf) == false){
			goto out;
		}
		memcpy(tmp_buf,buf,length);
		if(nf_writeblock(mtd,phy_page/this->ppb,(unsigned char *)tmp_buf) == false){
			goto out;
		}
        }
out:
	if(tmp_buf != NULL)
		kfree(tmp_buf);
	return ret;
}

void nf_bbt_collect_test(struct mtd_info *mtd,int block_addr)
{
	struct nand_chip *this = (struct nand_chip *)mtd->priv;
	int remap_block = 0,free_block = 0,bbt_len = 0;
	int find_phy_block = 0,remap_phy_block = nf_get_phyblock(mtd,block_addr);
	remap_block = remap_phy_block;
	printf("%s %d mark bad block log:%d phy:%d alway use remap fail case!\n",
		__FUNCTION__,__LINE__,block_addr,remap_block);
	/*get free current bbt number -->use the same block*/
	while(this->free_rba_num > 0)
	{
		/*reduce bbt number in the memory*/
		free_block = nf_get_freeblock(mtd, remap_block);
		printf("%s %d mark bad block %d  free block %d as bad block free bbt num %d "
			"use bbt num %d \n",__FUNCTION__,__LINE__,remap_block,free_block,
			this->free_rba_num,this->use_rba_num);
		remap_block = free_block;
	}
	dump_nf_bbt(mtd);
	find_phy_block = this->bbt[this->bbt_len - this->use_rba_num%this->bbt_len -1].remap_block;
	/*collect bbt region block*/
	printf("\n %s %d use collect bbt mode!can re-use block num:%d! \n \n",
		__FUNCTION__,__LINE__,this->use_rba_num);
	remap_block = remap_phy_block;
	bbt_len = this->bbt_len;
	while(bbt_len > 0){
		free_block = nf_get_freeblock(mtd,remap_block);
		if(free_block == 0xFFFFFFFF){
			printf("%s %d remap %d not enough free block \n",__FUNCTION__,
				__LINE__,remap_block);
			break;
		}
		printf("%s %d remap_block is %d free block %d \n",
	  		__FUNCTION__,__LINE__,remap_block,free_block);
		remap_block = free_block;
		bbt_len--;
	}
	printf("\n %s %d log block:%d phy block:%d  \n \n",
		__FUNCTION__,__LINE__,block_addr,find_phy_block);
	dump_nf_bbt(mtd);
}

static void nf_rbt_collect_test(struct mtd_info *mtd,int block_addr)
{
	struct nand_chip *this = (struct nand_chip *)mtd->priv;
	int remap_block = nf_get_phyblock(mtd,block_addr),rbt_len = 0,free_block = 0;
	int find_phy_block = 0;
	printf("%s %d mark bad block log:%d phy:%d,use different bad block!\n",
		__FUNCTION__,__LINE__,block_addr,remap_block);
	/*get free current bbt number -->use different logical block*/
	while(this->free_rba_num > 0)
	{
		/*reduce bbt in the memory*/
		free_block  = nf_get_freeblock(mtd, remap_block);
		printf("%s %d mark block_addr %d as bad block remap_block is %d "
			"free bbt num %d use bbt num %d \n",
			__FUNCTION__,__LINE__,remap_block,free_block,
			this->free_rba_num,this->use_rba_num);
		remap_block++;
	}
	dump_nf_bbt(mtd);
	/*get remap block*/
	remap_block = this->bbt[0].remap_block;
	/*this->use_rba_num-->collect number*/
	find_phy_block = this->bbt[this->use_rba_num -1].bad_block;
	printf("\n %s %d collect rbt mode! only collect block num:%d! \n \n",
		__FUNCTION__,__LINE__,this->use_rba_num);
	rbt_len = this->use_rba_num;
	while(rbt_len > 0)
	{
		/*reduce bbt in the memory*/
		free_block  = nf_get_freeblock(mtd, remap_block);
		if(free_block == 0xFFFFFFFF){
			printf("%s %d remap %d not enough free block \n",__FUNCTION__,__LINE__,
					remap_block);
			break;
		}
		printf("%s %d remap_block is %d free_block %d \n",
			__FUNCTION__,__LINE__,remap_block,free_block);
		rbt_len--;
	}
	printf("%s %d log block:%d phy block:%d \n",
		__FUNCTION__,__LINE__,block_addr,find_phy_block);
	dump_nf_bbt(mtd);
	dump_nf_rbt(mtd);
}

static void nf_table_test(struct mtd_info *mtd,int block_addr,int test_way)
{
	if(test_way == 0){
		nf_bbt_collect_test(mtd,block_addr); 
	}else if(test_way == 1){
		nf_rbt_collect_test(mtd,block_addr);
	}
}

void nf_test_erase_write_read(struct mtd_info *mtd,int block_addr)
{
#ifdef CONFIG_MTD_RBT_TEST
	struct nand_chip *this = (struct nand_chip *)mtd->priv;
	struct erase_info instr;
	unsigned long long to = 0;
	unsigned int  retlen = 0;
	u_char* write_buffer = NULL, *read_buffer = NULL;
	int loop = 0,page_id = 0,offset = 0,fin_item = 0,fin_phy_block = 0;
	instr.mtd = mtd;
	instr.addr = block_addr*this->block_size;
	instr.len = this->block_size;
	to = block_addr*this->block_size;
	if((write_buffer = kmalloc(this->block_size*sizeof(u_char),GFP_KERNEL)) == NULL){
		printf("%s %d malloc write_buffer error! \n",__FUNCTION__,__LINE__);
		return;
	}
	
	if((read_buffer = kmalloc(this->block_size*sizeof(u_char),GFP_KERNEL)) ==  NULL){
		printf("%s %d malloc read_buf error! \n",__FUNCTION__,__LINE__);
		kfree(write_buffer);
		return;
	}
	fin_item = ((1+2*this->ppb)*2)%this->bbt_len - 1;
	fin_phy_block = this->bbt[fin_item].remap_block;
	printf("%s %d make erase/write/read fail block num %d!\n",
		__FUNCTION__,__LINE__,(1+2*this->ppb)*2);
	nf_get_random_content(write_buffer,this->block_size);
	if(block_addr < this->bootcode_end_log_block){
		printf("%s %d this block_addr %d not kernel block %d \n",
			__FUNCTION__,__LINE__,block_addr,this->bootcode_end_log_block);
		return;
	}else{
		for(loop = 0; loop < 2 ; loop++)
		{
			test_erase_fail_remap = 1;
			printf("%s %d make erase block %d fail "
				"use bbt free bbt num %d,use bbt num %d!\n",
				__FUNCTION__,__LINE__,block_addr,
				this->free_rba_num,this->use_rba_num);
			mtd->erase(mtd, &instr);
			printf("%s %d make write or read  page fail total use block num %d !\n",
				__FUNCTION__,__LINE__,this->ppb*2);
			for(page_id = 0; page_id < this->ppb; page_id++)
			{
				offset = page_id*this->page_size;
				test_write_fail_remap = 1; /*this->ppb bad block*/
				printf("%s %d make write page %d fail "
					"use bbt free bbt num %d,use bbt num %d!\n",
					__FUNCTION__,__LINE__,block_addr*this->ppb+page_id,
					this->free_rba_num,this->use_rba_num);
				if(mtd->write_ecc(mtd, to+offset, this->page_size,&retlen, 
					(u_char *)write_buffer+offset,NULL, NULL)!= 0){
					printf("%s %d write fail page id %d\n", 
						__FUNCTION__,__LINE__,
						((int)(to+offset))/this->page_size);
					break;
				}
				test_read_warning_remap = 1; /*this->ppb bad block*/
				printf("%s %d make read page %d fail "
					"use bbt free bbt num %d,use bbt num %d!\n",
					__FUNCTION__,__LINE__,block_addr*this->ppb+page_id,
					this->free_rba_num,this->use_rba_num);
				mtd->read_ecc(mtd, to+offset, this->page_size, &retlen, (u_char *)read_buffer+offset, NULL, NULL);
				if(memcmp(write_buffer+offset,read_buffer+offset,this->page_size) == 0){
					printf("compare page_id %d offset 0x%x write read buffer ok! \n",page_id,offset);
				}else{
					printf("compare page_id %d offset 0x%x write read buffer fail! \n",page_id,offset);
				}	
			}
		}
	}
	
	printf("%s %d  log block:%d phy block:%d!\n",
		__FUNCTION__,__LINE__,block_addr,fin_phy_block);
	dump_nf_bbt(mtd);
	if(read_buffer!= NULL){
		kfree(read_buffer);
	}
	
	if(write_buffer!= NULL){
		kfree(write_buffer);
	}
#endif
}

void nf_test_remap(struct mtd_info *mtd,int block_addr,int way)
{
#ifdef CONFIG_MTD_RBT_TEST
	struct nand_chip *this = (struct nand_chip *)mtd->priv;
	struct erase_info instr;
	unsigned long long to = 0;
	unsigned int  retlen = 0;
	u_char* write_buffer = NULL,*read_buffer = NULL;
	int page_id = 0,offset = 0,compare_offset = 0,loop = 0,fin_phy_block = 0,fin_item = 0;
	instr.mtd = mtd;
	instr.len = this->block_size;

	if((write_buffer = kmalloc(this->block_size*sizeof(u_char),GFP_KERNEL)) == NULL){
		printf("%s malloc write_buffer error!\n",__FUNCTION__);
		return;
	}

	if((read_buffer = kmalloc(this->block_size*sizeof(u_char),GFP_KERNEL)) == NULL){
		printf("%s malloc read_buf error!\n",__FUNCTION__);
		kfree(write_buffer);
		return;
	}

	instr.addr = block_addr*this->block_size;
	if(mtd->erase(mtd, &instr) < 0){
		printf("%s erase block %d error!\n",__FUNCTION__,block_addr);
		return;
	}
	nf_get_random_content(write_buffer,this->block_size);
	to = block_addr*this->block_size;
	if(way == 0){
		fin_item = this->ppb%this->bbt_len - 1;
		printf("%s %d make read fail block num %d!\n",
			__FUNCTION__,__LINE__,this->ppb);
	}else{
		fin_item = (this->ppb/2)%this->bbt_len - 1;
		printf("%s %d make write fail block num %d!\n",
			__FUNCTION__,__LINE__,this->ppb/2);
	}
			
	fin_phy_block = this->bbt[fin_item].remap_block;
	/*way =0 -->test read way:write page 0/2/4...  page 1/3/5...no write 0xFF 
	   read all page  page 0/2/4...  just warning ok   page 1/3/5... read fail  all use bbt 
	    -->compare page 0/2/4... ok  page 1/3/5... compare 0xFF
	   way=1-->test write way:write page 0/2/4... use bbt  page 1/3/5...  write.
	   read all page  page
	   -->compare page all ok*/
	for(page_id = 0; page_id < this->ppb; page_id++)
	{
		offset = page_id*this->page_size;
		if(way == 0){
			if(page_id%2 == 0){
				if(mtd->write_ecc(mtd, to+offset, this->page_size, &retlen, 
					(u_char *)write_buffer+offset,NULL, NULL)!= 0){
					printf("%s %d write fail page id %d\n", 
						__FUNCTION__,__LINE__,page_id);
					break;
				}
				test_read_warning_remap = 1; /*read content ok,this->ppb/2bad block*/
			}else{
				test_read_fail_remap = 1; /*read content fail-->use 0xFF content*/
			}
			printf("%s %d read page %d use bbt free bbt num %d,use bbt num %d!\n",
				__FUNCTION__,__LINE__,block_addr*this->ppb+page_id,
					this->free_rba_num,this->use_rba_num);
		}else{
			if(page_id%2 != 0){
				test_write_fail_remap = 1;
				printf("%s %d make write page %d fail "
				"use bbt free bbt num %d,use bbt num %d!\n",
				__FUNCTION__,__LINE__,block_addr*this->ppb+page_id,
					this->free_rba_num,this->use_rba_num);
			}
			if(mtd->write_ecc(mtd, to+offset, this->page_size, &retlen, 
				(u_char *)write_buffer+offset,NULL, NULL)!= 0){
				printf("%s %d write fail page id %d\n", __FUNCTION__,__LINE__,page_id);
				break;
			}
		}
		
		mtd->read_ecc(mtd, to+offset, this->page_size, &retlen,
			(u_char *)read_buffer+offset, NULL, NULL);
		if(way == 0){
			memset((u_char *)read_buffer,0xFF,this->block_size);
			/*read again,make sure the pre data ok!*/
			mtd->read_ecc(mtd, to, this->page_size*(page_id+1), &retlen,
				(u_char *)read_buffer, NULL, NULL);
		}

		for(loop = 0; loop < page_id+1; loop++)
		{
			compare_offset = loop*this->page_size;
			printf("%s %d loop %d page id %d\n",__FUNCTION__,__LINE__,loop,page_id);
			nf_dump_byte((char *)(read_buffer+compare_offset), 16);
			if((loop%2 == 0)||(way == 1)){
				if(memcmp(write_buffer+compare_offset,read_buffer+compare_offset,this->page_size) == 0){
					printf("compare page_id %d offset 0x%x write read buffer ok! \n",loop,compare_offset);
				}else{
					printf("compare page_id %d offset 0x%x write read buffer fail! \n",loop,compare_offset);
				}
			}else{
				if(check_buffer_empty(read_buffer+compare_offset,this->page_size) == 0){
					printf("compare page_id %d offset 0x%x write read buffer ok! \n",loop,compare_offset);
				}else{
					printf("compare page_id %d offset 0x%x write read buffer fail! \n",loop,compare_offset);
				}
			}
		}
	}

	printf("%s %d  log block:%d phy block:%d!\n",
		__FUNCTION__,__LINE__,block_addr,fin_phy_block);
	dump_nf_bbt(mtd);
	if(read_buffer!= NULL){
		kfree(read_buffer);
	}
	
	if(write_buffer!= NULL){
		kfree(write_buffer);
	}
#endif
}

void nf_test_no_collect_table(struct mtd_info *mtd,int block_addr,int way)
{
#ifdef CONFIG_MTD_RBT_TEST
	struct nand_chip *this = (struct nand_chip *)mtd->priv;
	struct erase_info instr;
	unsigned long long to = 0;
	unsigned int  retlen = 0;
	u_char* write_buffer = NULL;
	int page_id = 0,offset = 0,remap_block = 0,bad_block = 0;
	instr.mtd = mtd;
	instr.len = this->block_size;
	instr.addr = block_addr*this->block_size;
	to = block_addr*this->block_size;
	if((write_buffer = kmalloc(this->block_size*sizeof(u_char),GFP_KERNEL)) == NULL){
		printf("%s %d malloc write_buffer error! \n",__FUNCTION__,__LINE__);
		return;
	}

	bad_block = block_addr;
	while(this->free_rba_num > 0)
	{
		/*reduce bbt number in the memory*/
		remap_block  = nf_get_freeblock(mtd, bad_block);
		if(way == 0){
			bad_block   = remap_block;
		}else{
			bad_block++;
		}
		printf("%s %d  block_add %d remap_block is %d  free bbt num %d "
			"cur bbt num %d \n",__FUNCTION__,__LINE__,bad_block,remap_block,
			this->free_rba_num,this->use_rba_num);
	}
	dump_nf_bbt(mtd);

	/*no free block -->use collect block -->way = 0 collect bbt  way =1 collect rbt*/
	if(mtd->erase(mtd, &instr) < 0){
		printf("%s erase block %d error! \n",__FUNCTION__,remap_block);
		return;
	}
	if(way == 0){
		printf("%s %d bbt test frist collect bbt then collect rbt,need max block %d!\n",
			__FUNCTION__,__LINE__,this->bbt_len+1);
	}else{
		printf("%s %d rbt test collect rbt only,need max block %d!\n",
			__FUNCTION__,__LINE__,this->bbt_len);
	}

	nf_get_random_content(write_buffer,this->block_size);
	for(page_id = 0; page_id < this->ppb; page_id++)
	{
		offset = page_id*this->page_size;
		/*make all collect block(this->bbt_len + 1),include bbt region(this->bi_table_block_num)+10
		  erase ok, write fail case, so no block to write ,so need break case*/
		test_write_fail_remap = this->bbt_len + 10; 
		printf("%s %d bbt or rbt collect log block %d phy block %d ok, write fail! "
			"write fail block num %d!\n",__FUNCTION__,__LINE__,
			block_addr,nf_get_phyblock(mtd,block_addr),test_write_fail_remap);
		if(mtd->write_ecc(mtd, to+offset, this->page_size, &retlen, (u_char *)write_buffer+offset,
			NULL, NULL)!= 0){
			printf("%s %d write fail page id %d\n", __FUNCTION__,__LINE__,page_id);
			break;
		}	
	}

	if(write_buffer!= NULL){
		kfree(write_buffer);
	}
#endif
}

void nf_test_bi_table(struct mtd_info *mtd)
{
#ifdef CONFIG_MTD_RBT_TEST
	struct nand_chip *this = (struct nand_chip *)mtd->priv;
	dump_nf_bi_table(mtd);
	printf("read page error,entry recovery flow,recovery read fail page!\n");
	test_read_fail_remap = 1;
	load_table_from_nand(mtd,0xBB,0);
	dump_nf_bi_table(mtd);
	printf("\n read page error,entry recovery flow,recovery write page fail"
		" and use free page instead!\n");
	test_read_fail_remap = 1;
	test_write_fail_bbt = 1;
	load_table_from_nand(mtd,0xBB,0);
	dump_nf_bi_table(mtd);
	printf("\n write frist page error,entry use free block way!\n");
	test_write_fail_bbt = 1;
	update_bbt_region_to_nand(mtd,(uchar *)this->bbt,(sizeof(BB_t)*this->bbt_len));
	dump_nf_bi_table(mtd);
#endif
} 

void nf_erase_flash(struct mtd_info *rtk_mtd)
{
	struct nand_chip *this = (struct nand_chip *)rtk_mtd->priv;
	int loop = 0;
	printf("%s %d erase number %d!\n",__FUNCTION__,__LINE__,this->block_num); 
	for(loop = 0; loop < this->block_num; loop++)
	{
		memset(this->g_oobbuf,0xFF,this->oob_size);
		/*skip read oob error*/
		if (this->read_oob(rtk_mtd, 0, loop*this->ppb, this->oob_size, this->g_oobbuf)){
			printf ("%s %d read_oob block %d page=%d failed!\n",
				__FUNCTION__,__LINE__,loop,loop*this->ppb);
			continue;
		}
		/*skip read oob error*/
		if((this->g_oobbuf[0] == 0)
		       ||((this->g_oobbuf[0] != BLOCK_CLEAN)
			&&(this->g_oobbuf[0] != BLOCK_OTHER_DATA)
			&&(this->g_oobbuf[0] != BLOCK_IBT) 
			&&(this->g_oobbuf[0] != BLOCK_BBT)
			&&(this->g_oobbuf[0] != BLOCK_HWSETTING)
			&&(this->g_oobbuf[0] != BLOCK_KEY_SIG) 
			&&(this->g_oobbuf[0] != BLOCK_EMCU_CERT)
			&&(this->g_oobbuf[0] != BLOCK_EMCU)
			&&(this->g_oobbuf[0] != BLOCK_KBOOT)
			&&(this->g_oobbuf[0] != BLOCK_AUXCODE) 
			&&(this->g_oobbuf[0] != BLOCK_DATA) 
			&&(this->g_oobbuf[0] != BLOCK_ENVPARAM_MAGICNO) 
			&&(this->g_oobbuf[0] != BLOCK_FACTORY_SETTING))){
			printf ("%s %d read_oob block %d page=%d tag 0x%x failed!\n", 
				__FUNCTION__,__LINE__,loop,loop*this->ppb,this->g_oobbuf[0]);
			continue;
		}
		/*erase block*/
		if (this->erase_block (rtk_mtd, 0, loop*this->ppb) == 0){
			printf("%s %d erase blockID %d sucessful!\n",
				__FUNCTION__,__LINE__,loop); 
		}else{
			printf("%s %d erase blockID %d failed!\n",
				__FUNCTION__,__LINE__,loop); 
			mark_as_badblock(rtk_mtd,loop);
		}
	}
}