#include <common.h>
#include <asm/arch/system.h>
#include <drivers/otp/rtk_otp_debug.h>
#include <drivers/otp/rtk_otp_reg.h>
#include <drivers/otp/rtk_otp_hw.h>

#define OTP_BIT_MASK(i)       (0xFFFFFFFF >> (32-(i)))


/*-----------------------------------------------------------------------------
 * Func : _otp_get_bits
 *
 * Desc : get specified bits
 *
 * Parm :
 *
 * Retn :
 *-----------------------------------------------------------------------------*/
static int _otp_get_bits(unsigned long* pval, unsigned char msb, unsigned char lsb)
{
    if (msb < lsb || lsb > 31 || msb > 31)
        return -1;

    *pval >>= lsb;
    *pval &= OTP_BIT_MASK(msb - lsb + 1);

    return 0;
}




/*-----------------------------------------------------------------------------
 * Func : otp_read
 *
 * Desc : read OTP word
 *
 * Parm : ofst : byte offsets, it should be 4 bytes alignments
 *
 * Retn : OTP words
 *-----------------------------------------------------------------------------*/
int otp_read(unsigned long ofst, unsigned long *pval)
{
    return rtk_otp_hw_read(ofst, pval);
}


/*-----------------------------------------------------------------------------
 * Func : otp_read_bits
 *
 * Desc : read OTP bits
 *
 * Parm : ofst : byte offsets, it should be 4 bytes alignments
 *
 * Retn : OTP words
 *-----------------------------------------------------------------------------*/
int otp_read_bits(unsigned long ofst, unsigned char msb,
                                    unsigned char lsb, unsigned long *pval)
{
    int ret = otp_read(ofst, pval);

    if (ret == 0 && *pval != 0xdeaddead) // do not mask bits if it is non readable
        ret = _otp_get_bits(pval, msb, lsb);

    return ret;
}

/*-----------------------------------------------------------------------------
 * Func : otp_read_region
 *
 * Desc : read OTP region
 *
 * Parm : ofst : byte offsets, it should be 4 bytes alignments
 *
 * Retn : OTP words
 *-----------------------------------------------------------------------------*/
int otp_read_region(unsigned long ofst, unsigned int *data, int len)
{
    int i = 0;
    if (ofst & 0x3) {
        OTP_WARNING("read otp region failed, ofst %08lx is not 4 bytes alignments\n", ofst);
        return -1;
    }
    if (data == NULL || len < 0)
        return -1;
    for (i = 0; i < len; i++) {
        if (otp_read((unsigned int)ofst, (unsigned long *)data) != 0) {
            return -1;
        }
        ofst += sizeof(unsigned int);
        data++;
    }
    return 0;
}

/*-----------------------------------------------------------------------------
 * Func : otp_read_region_byte
 *
 * Desc : read OTP region
 *
 * Parm : ofst : byte offsets, it should be 4 bytes alignments
 *
 * Retn : OTP words
 *-----------------------------------------------------------------------------*/
int otp_read_region_byte(unsigned long ofst, unsigned char *data, int len)
{
    unsigned char *tmp = data;
    if (tmp == NULL || len < 0) {
        OTP_WARNING("otp program region failed, params are input error\n");
        return -1;
    }
    /*if ofst is not align by 4 bytes,should seperate read the ualigned parts*/
    if (ofst & 0x3 && len > 0) {
        unsigned int val = 0;
        unsigned char start_byte = (ofst & 0x3);
        unsigned char read_len = ((len >= sizeof(unsigned int) - start_byte)) ? (sizeof(unsigned int) - start_byte) : len;
        otp_read((unsigned int )(ofst - start_byte), (unsigned long *)&val);
        ofst += read_len;
        len -= read_len;
        val = val >> (start_byte * 8);
        while (read_len > 0) {
            *tmp = val & 0xFF;
            val >>= 8;
            tmp++;
            read_len--;
        }
        if(len <= 0)
            return 0;

    }

    if (len / sizeof(int) > 0) {
        int read_len = len / sizeof(int);
        unsigned int val = 0;
        while (read_len > 0) {
            unsigned char times = 4;
            otp_read((unsigned int)ofst, (unsigned long *)&val);
            while (times > 0) {
                *tmp = val & 0xFF;
                tmp++;
                val >>= 8;
                times --;
            }
            ofst += sizeof(unsigned int);
            read_len--;
        }
        len = len - (len / sizeof(int)) * sizeof(int);
    }

    if (len > 0) {
        unsigned int val = 0;
        otp_read((unsigned int)ofst, (unsigned long *)&val);
        while(len > 0) {
            *tmp = val & 0xFF;
            val >>= 8;
            tmp++;
            ofst += 1;
            len --;
        }

    }
    return 0;


}


/*-----------------------------------------------------------------------------
 * Func : otp_write
 *
 * Desc : write OTP words
 *
 * Parm : ofst : byte offsets, it should be 4 bytes alignments
 *        val  : value of OTP words
 *
 * Retn :
 *-----------------------------------------------------------------------------*/
int otp_write(unsigned long ofst, unsigned long val)
{
    return rtk_otp_hw_program(ofst, val);
}

/*-----------------------------------------------------------------------------
 * Func : otp_write_bits
 *
 * Desc : write OTP bits
 *
 * Parm : ofst : byte offsets, it should be 4 bytes alignments
 *        val  : value of OTP words
 *
 * Retn : OTP words
 *-----------------------------------------------------------------------------*/
int otp_write_bits(unsigned long ofst, unsigned char msb,
                                    unsigned char lsb, unsigned long val)
{
    if (msb < lsb || lsb > 31 || msb > 31)
        return -1;

    val &= OTP_BIT_MASK(msb - lsb + 1);
    val <<= lsb; 

    return otp_write(ofst, val);
}

/*-----------------------------------------------------------------------------
 * Func : otp_program_region
 *
 * Desc : opt program region
 *
 * Parm : ofst: the offset of the first reg of the program region you want
 *           data:the data array
 *           len:  the length of the data array
 *
 * Retn : OTP words
 *-----------------------------------------------------------------------------*/

int otp_program_region(unsigned long ofst, unsigned int *data, int len)
{
    int i = 0;
    if (ofst & 0x3) {
        OTP_WARNING("otp program region failed, ofst %lx is not 4 bytes alignments\n", ofst);
        return -1;
    }
    if (data == NULL || len < 0) {
        OTP_WARNING("otp program region failed, params are input error\n");
        return -1;
    }
    for (i = 0; i < len; i++) {
        OTP_DBG("REG_ADDR = 0x%08lx, DATA = 0x%08x\n", ofst + OTP_DDR_MAP_BASE, *data);
        if(otp_write(ofst, *data) == -1)
            return -1;
        ofst += sizeof(unsigned int);
        data++;
    }
    return 0;


}


/*-----------------------------------------------------------------------------
 * Func : otp_program_region_byte
 *
 * Desc : opt otp_program_region_byte region
 *
 * Parm : ofst: the offset of the first reg of the program region you want
 *           data:the data array
 *           len:  the length of the data array
 *
 * Retn : OTP words
 *-----------------------------------------------------------------------------*/

int otp_program_region_byte(unsigned long ofst, unsigned char *data, int len)
{
    unsigned char *tmp = data;
    if (tmp == NULL || len < 0) {
        OTP_WARNING("otp program region failed, params are input error\n");
        return -1;
    }
    /*if ofst is not align by 4 bytes,should seperate read the ualigned parts*/
    if (ofst & 0x3 && len > 0) {
        unsigned int val = 0;
        unsigned char start_byte = (ofst & 0x3);
        unsigned char write_len = ((len >= sizeof(unsigned int) - start_byte)) ? (sizeof(unsigned int) - start_byte) : len;
        unsigned char times = write_len;
        while (times > 0) {
            val = (val << 8) | *(tmp + times - 1);
            times--;
        }
        val = val << (start_byte * 8);
        OTP_DBG("REG_ADDR = 0x%08lx, DATA = 0x%08x\n", ofst - start_byte + OTP_DDR_MAP_BASE, val);
        if(otp_write(ofst - start_byte, val) == -1)
            return -1;
        tmp += write_len;
        ofst += write_len;
        len -= write_len;
        if(len <= 0)
            return 0;

    }

    if (len / sizeof(int) > 0) {
        int write_len = len / sizeof(int);
        while (write_len > 0) {
            unsigned int val = 0;
            unsigned char times = sizeof(unsigned int);
            while (times > 0) {
                val = (val << 8) | *(tmp + times - 1);
                times --;
            }
            OTP_DBG("REG_ADDR = 0x%08lx, DATA = 0x%08x\n", ofst + OTP_DDR_MAP_BASE, val);
            if(otp_write(ofst, val) == -1)
                return -1;
            tmp += sizeof(unsigned int);
            ofst += sizeof(unsigned int);
            len -= sizeof(unsigned int);
            write_len--;
        }

    }

    if (len > 0) {
        unsigned int val = 0;
        unsigned char *revert = tmp + len - 1;
        while (revert >= tmp) {
            val = (val << 8) | *revert;
            revert--;
        }
        OTP_DBG("REG_ADDR = 0x%08lx, DATA = 0x%08x\n", ofst + OTP_DDR_MAP_BASE, val);
        if(otp_write(ofst, val) == -1)
            return -1;
        ofst += len;
        tmp += len;
        len = 0;
    }

    return 0;


}

/*-----------------------------------------------------------------------------
 * Func : otp_crc_auto_region
 *
 * Desc : opt crc auto region
 *
 * Parm : pval:pval:the address of the crc value
 *
 * Retn : OTP words
 *-----------------------------------------------------------------------------*/
int otp_crc_auto_region(unsigned long *pval)
{
    return rtk_otp_hw_crc(0x0, MAX_OTP_OFST, pval);
}


/*-----------------------------------------------------------------------------
 * Func : otp_crc_partial_region
 *
 * Desc : otp_crc_partial_region
 *
 * Parm : start_ofst:the offset of start reg you want to verify.
 *      end_ofst:the offset of end reg you want to verify.
 *      pval:the address of the crc value
 * Retn : OTP words
 *-----------------------------------------------------------------------------*/
int otp_crc_partial_region(unsigned long start_ofst, unsigned long len, unsigned long *pval)
{
    return rtk_otp_hw_crc(start_ofst, len, pval);
}


